#include <windows.h>
#include <stdint.h>
#include <stddef.h>
#include <stdio.h>
#include <math.h>

enum
{
    PLAYER_ANIM_WALK_FORWARD        = 0x04,
    PLAYER_ANIM_WALK_FORWARD_1      = 0x05,
    PLAYER_ANIM_RUN_FORWARD         = 0x06,
    PLAYER_ANIM_RUN_FORWARD_1       = 0x07,
    PLAYER_ANIM_WALK_BACKWARD       = 0x08,
    PLAYER_ANIM_WALK_BACKWARD_1     = 0x09,
    PLAYER_ANIM_RUN_RIGHT           = 0x10,
    PLAYER_ANIM_RUN_RIGHT_1         = 0x11,
    PLAYER_ANIM_ROTATE_RIGHT        = 0x18,
    PLAYER_ANIM_ROTATE_RIGHT_1      = 0x19,
    PLAYER_ANIM_IDLE                = 0x22,
    PLAYER_ANIM_IDLE_0              = 0x23,
    PLAYER_ANIM_FALL                = 0x2C,
    PLAYER_ANIM_FALL_0              = 0x2D,
    PLAYER_ANIM_FALL_ALT            = 0x2E,
    PLAYER_ANIM_FALL_ALT_0          = 0x2F,
    PLAYER_ANIM_KICK                = 0x30,
    PLAYER_ANIM_KICK_0              = 0x31,
    PLAYER_ANIM_SQASH               = 0x32,
    PLAYER_ANIM_SQASH_0             = 0x33,
    PLAYER_ANIM_AIM                 = 0x38,
    PLAYER_ANIM_AIM_0               = 0x39,
    PLAYER_ANIM_AIM_STOP            = 0x42,
    PLAYER_ANIM_AIM_STOP_0          = 0x43,
    PLAYER_ANIM_DEAD                = 0x50,
    PLAYER_ANIM_DEAD_0              = 0x51,
    PLAYER_ANIM_STAND               = 0x68,
    PLAYER_ANIM_STAND_0             = 0x69,
    PLAYER_ANIM_SQASH_ALT           = 0x70,
    PLAYER_ANIM_SQASH_ALT_0         = 0x71
};

typedef struct player_s
{
    int16_t animation;

    int16_t _b, _c, _d, _e, _f, _g, _h, _i;

    struct { int32_t x, y, z; } p;
    struct { int16_t x, y, z; } r;

} player_t;

typedef struct camera_s
{
    struct { int32_t x, y, z; } p;

    int32_t _a, _b, _c, _d, _e, _f, _g, _h, _i, _j, _k, _l, _m, _n, _o, _p, _q, _r, _s, _t, _u, _v, _w;

    struct { int16_t x, y, z; } r;

} camera_t;

player_t player;

camera_t camera;

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
    /* Find ePSXe window */
    HWND hEpsxeWnd = FindWindow("EPSX", NULL);

    if (hEpsxeWnd) {
        /* Obtain process ID */
        DWORD epsxePID;
        GetWindowThreadProcessId(hEpsxeWnd, &epsxePID);
        if (epsxePID) {
            HANDLE hEpsxeProc = OpenProcess(PROCESS_ALL_ACCESS, FALSE, epsxePID);

            while (1) {
                if (GetAsyncKeyState(VK_END)) break;

                const size_t baseAddress = 0x1060000; /* todo: find a way to obtain this value automatically */
                const size_t playerStructOffset = baseAddress + 0xB3B620;
                const size_t cameraStructOffset = baseAddress + 0xB3B330;

                ReadProcessMemory(hEpsxeProc, (LPCVOID)playerStructOffset, &player, sizeof(player_t), NULL);
                ReadProcessMemory(hEpsxeProc, (LPCVOID)cameraStructOffset, &camera, sizeof(camera_t), NULL);

                double t = GetTickCount() / 1000.0;
                static const double movementSpeed = 70;
                static const double rotationSpeed = 5;

                /* Reset */
                if (GetAsyncKeyState(VK_HOME)) {
                    if (GetAsyncKeyState(VK_CONTROL)) {
                        /* Reset rotation */
                        ZeroMemory(&player.r, sizeof(player.r));
                    } else {
                        /* Reset rotation */
                        ZeroMemory(&player.p, sizeof(player.p));
                    }
                }

                /* Z movement */
                if (GetAsyncKeyState(VK_ADD))           { player.p.y -= movementSpeed; }
                if (GetAsyncKeyState(VK_SUBTRACT))      { player.p.y += movementSpeed; }

                if (GetAsyncKeyState(VK_LCONTROL)) {
                    /* X/Y/Z rotation */
                    if (GetAsyncKeyState(VK_NUMPAD8))   { player.r.x -= rotationSpeed; }
                    if (GetAsyncKeyState(VK_NUMPAD5))   { player.r.x += rotationSpeed; }
                    if (GetAsyncKeyState(VK_NUMPAD7))   { player.r.y -= rotationSpeed; }
                    if (GetAsyncKeyState(VK_NUMPAD9))   { player.r.y += rotationSpeed; }
                    if (GetAsyncKeyState(VK_NUMPAD4))   { player.r.z -= rotationSpeed; }
                    if (GetAsyncKeyState(VK_NUMPAD6))   { player.r.z += rotationSpeed; }
                } else {
                    if (GetAsyncKeyState(VK_MENU)) {
                        /* X/Y movement */
                        if (GetAsyncKeyState(VK_NUMPAD4))   { player.p.x -= movementSpeed; }
                        if (GetAsyncKeyState(VK_NUMPAD6))   { player.p.x += movementSpeed; }
                        if (GetAsyncKeyState(VK_NUMPAD5))   { player.p.z -= movementSpeed; }
                        if (GetAsyncKeyState(VK_NUMPAD8))   { player.p.z += movementSpeed; }
                        /* Y rotation */
                        if (GetAsyncKeyState(VK_NUMPAD7))   { player.r.y -= rotationSpeed; }
                        if (GetAsyncKeyState(VK_NUMPAD9))   { player.r.y += rotationSpeed; }
                    } else {
                        /* Y rotation */
                        if (GetAsyncKeyState(VK_NUMPAD7))   { player.r.y -= rotationSpeed; }
                        if (GetAsyncKeyState(VK_NUMPAD9))   { player.r.y += rotationSpeed; }
                        /* Relative movement */
                        double rdeg = ((double)player.r.y / (1024 * 4)) * 360;
                        double rcos = cos(rdeg * (M_PI / 180));
                        double rsin = sin(rdeg * (M_PI / 180));
                        if (GetAsyncKeyState(VK_NUMPAD8)) {
                            player.p.z += rcos * movementSpeed;
                            player.p.x += rsin * movementSpeed;
                        }
                        if (GetAsyncKeyState(VK_NUMPAD5)) {
                            player.p.z -= rcos * movementSpeed;
                            player.p.x -= rsin * movementSpeed;
                        }
                        if (GetAsyncKeyState(VK_NUMPAD4)) {
                            player.p.z += rsin * movementSpeed;
                            player.p.x -= rcos * movementSpeed;
                        }
                        if (GetAsyncKeyState(VK_NUMPAD6)) {
                            player.p.z -= rsin * movementSpeed;
                            player.p.x += rcos * movementSpeed;
                        }
                        /* 3'rd person camera */
                        camera.p.x = player.p.x - 8000;
                        camera.p.y = player.p.y - 8000;
                        camera.p.z = player.p.z - 8000;
                        camera.r.x = player.r.x;
                        camera.r.y = player.r.y;
                        camera.r.z = 0;
                    }
                }

                WriteProcessMemory(hEpsxeProc, (LPVOID)playerStructOffset, &player, sizeof(player_t), NULL);
                WriteProcessMemory(hEpsxeProc, (LPVOID)cameraStructOffset, &camera, sizeof(camera_t), NULL);

                const size_t cameraFovOffset = baseAddress + 0xB3E250;
                uint16_t fov;
                ReadProcessMemory(hEpsxeProc, (LPCVOID)cameraFovOffset, &fov, sizeof(fov), NULL);
                fov = 254;
                WriteProcessMemory(hEpsxeProc, (LPVOID)cameraFovOffset, &fov, sizeof(fov), NULL);

                Sleep(1);
            }
        } else {
            WriteConsole(GetStdHandle(STD_OUTPUT_HANDLE), "Can't obtain process!", 0, NULL, NULL);
        }
    } else {
        WriteConsole(GetStdHandle(STD_OUTPUT_HANDLE), "EPSXE window not found!", 0, NULL, NULL);
    }
    return 0;
}
